@@ -25,6 +25,7 @@
 #include "cspi/spi.h"
 #include "SRObject.h"
 #include "string.h"
+#include  "gdk/gdk.h"
 
 #define	SRL_MAX_CLIENTS			1
 #define	SRL_LOG
@@ -93,7 +94,7 @@ typedef enum _SRLEventType
     SRL_EVENT_TEXT_SELECTION_CHANGED,
     SRL_EVENT_VALUE_CHANGED,
     SRL_EVENT_STATE_CHECKED,
-    SRL_EVENT_STATE_SELECTED, /*FIXME: really needed???? */
+    SRL_EVENT_STATE_SELECTED, 
     SRL_EVENT_STATE_EXPANDED,
     SRL_EVENT_STATE_SHOWING,
     SRL_EVENT_SELECTION_CHANGED,
@@ -125,6 +126,7 @@ typedef enum _SRLEventPriority
     SRL_EVENT_PRIORITY_MOUSE,
     SRL_EVENT_PRIORITY_TOOLTIP,
     SRL_EVENT_PRIORITY_FOCUS,
+    SRL_EVENT_PRIORITY_WATCHED,
     SRL_EVENT_PRIORITY_WINDOW,
     SRL_EVENT_PRIORITY_LAST,
     SRL_EVENT_PRIORITY_UNKNOWN = SRL_EVENT_PRIORITY_LAST
@@ -734,13 +736,29 @@ srl_is_window_event (SRLEvent *event)
 	    event->type == SRL_EVENT_TAB_ADDED ||
 	    event->type == SRL_EVENT_TAB_REMOVED;
 }
+static gboolean
+srle_is_for_watched_acc (SRLEvent *event)
+{
+    srl_assert (event);
+    
+    return srl_watched_acc == event->acc_ev->source;
+}
 
+static gboolean
+srle_is_for_focused_acc (SRLEvent *event)
+{
+    srl_assert (event);
+
+    return srl_last_focus == srle_get_acc (event);
+}
+
+/*
 static gboolean
 srl_is_focus_event (SRLEvent *event)
 {
     srl_assert (event);
     
-    return  event->type == SRL_EVENT_FOCUS		||
+    return  (event->type == SRL_EVENT_FOCUS		||
 	    event->type == SRL_EVENT_FOCUS1		||
 	    event->type == SRL_EVENT_FOCUS2		||
 	    event->type == SRL_EVENT_LINK_SELECTED	||
@@ -755,7 +773,33 @@ srl_is_focus_event (SRLEvent *event)
 	    event->type == SRL_EVENT_STATE_EXPANDED	||
 	    event->type == SRL_EVENT_STATE_SHOWING	||
 	    event->type == SRL_EVENT_SELECTION_CHANGED	||
-	    event->type == SRL_EVENT_VISIBLE_DATA_CHANGED;
+	    event->type == SRL_EVENT_VISIBLE_DATA_CHANGED) &&
+	    srle_is_for_focused_acc (event);
+}
+*/
+
+static gboolean
+srl_is_watched_event (SRLEvent *event)
+{
+    srl_assert (event);
+    
+    return  (event->type == SRL_EVENT_FOCUS		||
+	    event->type == SRL_EVENT_FOCUS1		||
+	    event->type == SRL_EVENT_FOCUS2		||
+	    event->type == SRL_EVENT_LINK_SELECTED	||
+	    event->type == SRL_EVENT_TEXT_CARET_MOVED	||
+	    event->type == SRL_EVENT_TEXT_CHANGED_INSERT||
+	    event->type == SRL_EVENT_TEXT_CHANGED_DELETE||
+	    event->type == SRL_EVENT_TEXT_SELECTION_CHANGED||
+	    event->type == SRL_EVENT_VALUE_CHANGED	||
+	    event->type == SRL_EVENT_NAME_CHANGED	||
+	    event->type == SRL_EVENT_STATE_CHECKED	||
+	    event->type == SRL_EVENT_STATE_SELECTED	||
+	    event->type == SRL_EVENT_STATE_EXPANDED	||
+	    event->type == SRL_EVENT_STATE_SHOWING	||
+	    event->type == SRL_EVENT_SELECTION_CHANGED	||
+	    event->type == SRL_EVENT_VISIBLE_DATA_CHANGED) &&
+	    srle_is_for_watched_acc (event);
 }
 
 static gboolean
@@ -778,17 +822,16 @@ static SRLEventPriority
 srle_get_priority (SRLEvent *event)
 {
     srl_assert (event);
-    
-    if (srl_is_focus_event (event))
-	return SRL_EVENT_PRIORITY_FOCUS;
+	
+    if (srl_is_watched_event (event))
+	return SRL_EVENT_PRIORITY_WATCHED;
     if (srl_is_window_event (event))
 	return SRL_EVENT_PRIORITY_WINDOW;
     if (srl_is_mouse_event (event))
 	return SRL_EVENT_PRIORITY_MOUSE;
     if (srl_is_tooltip_event (event))
 	return SRL_EVENT_PRIORITY_TOOLTIP;
-    srl_assert_not_reached ();
-    return SRL_EVENT_PRIORITY_UNKNOWN;
+    return SRL_EVENT_PRIORITY_FOCUS;
 }
 
 static gboolean
@@ -1047,6 +1090,8 @@ srl_report_event_to_clients (gpointer data)
 {
     srl_report_obj_event_to_clients (SRL_EVENT_PRIORITY_WINDOW, 
 					    SR_EVENT_WINDOW);    
+    srl_report_obj_event_to_clients (SRL_EVENT_PRIORITY_WATCHED,
+					    SR_EVENT_SRO);
     srl_report_obj_event_to_clients (SRL_EVENT_PRIORITY_FOCUS,
 					    SR_EVENT_SRO);
     srl_report_obj_event_to_clients (SRL_EVENT_PRIORITY_TOOLTIP,
@@ -1147,14 +1192,6 @@ srl_acc_has_toolkit (const Accessible *acc,
     return rv;    
 }
 
-static gboolean
-srle_is_for_watched_acc (SRLEvent *event)
-{
-    srl_assert (event);
-    
-    return srl_watched_acc == event->acc_ev->source;
-}
-
 static gboolean
 srle_is_for_focused_ancestor_acc (SRLEvent *event)
 {
@@ -1180,14 +1217,6 @@ srle_is_for_focused_ancestor_acc (SRLEvent *event)
 }
 
 
-static gboolean
-srle_is_for_focused_acc (SRLEvent *event)
-{
-    srl_assert (event);
-
-    return srl_last_focus == srle_get_acc (event);
-}
-
 static gboolean
 srl_event_is_for_metacity (SRLEvent *event)
 {
@@ -1278,6 +1307,16 @@ srle_change_type (SRLEvent *event)
 	event->type = SRL_EVENT_TAB_ADDED;
     else if (srle_has_type (event, SRL_EVENT_CHILDREN_REMOVED))
 	event->type = SRL_EVENT_TAB_REMOVED;
+	
+    /* Convert the "state-changed:selected" event for a _selected_ page-tab
+       into a "focus:" event */
+    if (srle_has_type (event, SRL_EVENT_STATE_SELECTED) &&
+        srle_acc_has_role (event, SPI_ROLE_PAGE_TAB)    &&
+	srl_acc_has_state (event->acc_ev->source, SPI_STATE_SELECTED))
+    {
+	event->type = SRL_EVENT_FOCUS;
+    }	       	
+    
     return TRUE;
 }
 
@@ -1612,6 +1651,18 @@ srl_event_is_reported (SRLEvent *event)
          srle_has_type (event, SRL_EVENT_TAB_REMOVED))) 
 	rv = srle_acc_has_role (event, SPI_ROLE_PAGE_TAB);
 	
+    /*In some situations we have two 'focus:' events for a page-tab 
+     (an "object:state-changed:selected" event which was converted into
+      a focus event _and_ a "focus:" event) => report a single "focus:" event*/	
+    if (srle_has_type (event, SRL_EVENT_FOCUS) &&
+        srle_acc_has_role (event, SPI_ROLE_PAGE_TAB))
+    {
+	if (srl_last_focus == event->acc_ev->source)
+	    rv = FALSE;
+	else
+	    rv = TRUE;    
+    }	      
+	
     return rv;
 }
 
@@ -1636,6 +1687,7 @@ srl_process_event (SRLEvent *event)
     srl_timing_process_start ((gpointer) event->acc_ev);
     srle_set_acc (event);
     srle_change_type (event);
+    process = srl_event_is_reported (event);
 
     if (srle_has_type (event, SRL_EVENT_FOCUS) ||
 	srle_has_type (event, SRL_EVENT_FOCUS2))
@@ -1646,7 +1698,6 @@ srl_process_event (SRLEvent *event)
 /* FIXME: 
     if (srle_has_type (event, SRL_EVENT_WINDOW_DEACTIVATE))
 	report no focus if no other event */
-    process = srl_event_is_reported (event);
     if (process)
 	srl_report_event (event);
 
@@ -1734,18 +1785,20 @@ srl_init ()
 	    rv = SPI_registerGlobalEventListener (srl_event_listeners[i],
 				srl_events_type_name[i].name);
 	    if (!rv)
-	    	srl_error ("Cannot register a listener for event \"%s\".",
+	    {
+		gdk_beep ();
+	    	srl_warning ("Cannot register a listener for event \"%s\".",
 				srl_events_type_name[i].name);
+	    }
 	}
 	else
 	{
-	    srl_error ("Cannot create a listener for event \"%s\"",
+	    srl_warning ("Cannot create a listener for event \"%s\"",
 				srl_events_type_name[i].name);		
-	    rv = FALSE;
+	    rv = TRUE;
 	}
-	if (!rv)
-	    break;
     }
+    rv = TRUE;
 
     if (rv)
 	srl_set_initialized ();
